#!/usr/bin/env python
# -*- coding: utf-8 -*-

# import
import logging

import os
import time
import netCDF4
from pylab import date2num, num2date#, flatten
from numpy import array, arange, float32, NaN, transpose, int16, diff, logical_and, meshgrid, ravel, zeros, unique, nanmean, isnan, hstack
from ReadRawFile import ReadRawFile
from scipy import spatial
from mpl_toolkits.basemap import pyproj

# ---Class for satellite name--- #
class SatName:
    """Defined a number for each satellite"""
    def __init__(self):
        self.satname = array(['AMSUBN15', 'AMSUBN16', 'MHSN18', 'MHSN19', 'MHSM01', 'MHSM02'])
        self.satnum = arange(self.satname.size)+1

    def get_satnum(self, name):
        return array([self.satnum[nn == self.satname] for nn in name])
        # return concatenate([self.satnum[nn == self.satname] for nn in name])

    def get_satname(self, num):
        return array([self.satname[nn == self.satnum] for nn in num])
        # return concatenate([self.satname[nn == self.satnum] for nn in num])


# ---Correction for one orbit--- #
def applycorr2orb(fname, lonvec, latvec, ch=['All'], area=[-35., 35., -180., 180.]):
    """Apply correction to AMSU or MHS orbit
    *Inputs:
    fname  = file name
    tstep  = time resolution for new grid in minutes
    lonvec = longitude vecteur of new grid
    latvec = latitude vecteur of new grid
    ch     = list of extrated channel ('B1', 'B2', 'B3', 'B4', 'B5' or 'All'), default 'All'
    area   = limits [LS, LN, LW, LE] for extracted area, default is tropics (ie [-35., 35., -180., 180.])

    *Outputs:
    tbcorrm = brightness temperature of orbit in new grid
    trefm   = ordinal time
    """

    t = time.time()

    logger = logging.getLogger()

    if ch[0].lower() == 'all':
        ch = ['B1', 'B2', 'B3', 'B4', 'B5']

    # Read orbit file
    lon, lat, dt, tb, fov = load_amsub_mhs(fname, ch, area)

    try:
        # Coef of correction as a function of classification, fov and day of year
        coefz = find_corrcoef(dt, lon, lat, fov, ch)
        nonan = ~isnan(coefz[0,:])

    except StandardError:
        logger.error('Orbit is out of the area: {0}'.format(fname))
        return [], [], []

    else:
        tbcorr = array(tb)[:,nonan] - coefz[:,nonan]
        # Average on new grid
        tbcorrm, tobsm = newgrid(lonvec, latvec, lon[nonan], lat[nonan], dt[nonan], tbcorr)

    logger.info('Deal {0} in: {1}'.format(fname, time.time()-t))

    return tbcorrm, tobsm, fname.split('/')[3]

# ---Read AMSU-B and MHS files--- #
def load_amsub_mhs(fname, ch, area):
    """Read AMSU-B and MHS files
    *Inputs:
    fname = file name
    ch    = list of extrated channel ('B1', 'B2', 'B3', 'B4', 'B5')
    area  = limits [LS, LN, LW, LE] for extracted area, default is Tropics (ie [-35, 35, -180., 180])

    *Outputs:
    lon  = Longitude
    lat  = Latitude
    yday = Day of year
    tms  = UTC time in milliseconds
    tb   = Brightness temperature for each channel
    err  = 1 if reading error, 0 else
    """

    # Index channel to extract
    ich = array([int(cch[1])-1 for cch in ch])

    project = os.environ['PROJECT']
    hdr_file = '{0}/src/external_python/ReadRawFile/data/amsub-mhs_1c_header'.format(project)
    rec_file = '{0}/src/external_python/ReadRawFile/data/amsub-mhs_1c_record'.format(project)
    # Variables
    dset_names = ['amb1c_latlon', 'amb1c_btemps', 'amb1c_scnlindy', 'amb1c_scnlintime', 'amb1c_scnlinyr']
    try:
        headers, datasets = ReadRawFile(fname, hdr_file, rec_file, False, dset_names)   
    except:
        logger.error('Error with {0}'.format(fname))
        # continue
        return

    # Latitude
    lat = float32(datasets['amb1c_latlon'][:, :, 0]/1.E4)
    # Longitude
    lon = float32(datasets['amb1c_latlon'][:, :, 1]/1.E4)
    # Year
    yyyy = datasets['amb1c_scnlinyr']
    # Day of Year
    yday = datasets['amb1c_scnlindy']
    # Time in milliseconds
    tms = datasets['amb1c_scnlintime']
    # Tb
    tb = float32(datasets['amb1c_btemps'][:, :, ich]/1.E2)
    # Remplace fill values -9999.99 by NaN
    tb[(tb == -9999.99)] = NaN

    # Extract Area
    # Index in the defined area
    LS, LN, LW, LE = area
    zone = (lon >= LW) & (lon <= LE) & (lat >= LS) & (lat <= LN)
    nscn, nfov, nch = tb.shape
    # Extraction of Tb for each channel
    tb = [(tb[:, :, ii])[zone] for ii in xrange(nch)]
    # day of year and time
    yyyy = transpose([yyyy]*nfov)[zone]
    yday = transpose([yday]*nfov)[zone]
    tms = transpose([tms]*nfov)[zone]
    # Find location in fov
    fov = array(range(nfov)*nscn).reshape(nscn, nfov)[zone]+1

    # Convert in python date/time
    dd = num2date(yday)
    tt = [ms2time(ms) for ms in tms]
    dt = array([d.replace(year=y, hour=h, minute=m, second=s) for d, y, (h, m, s) in zip(dd, yyyy, tt)])

    return lon[zone], lat[zone], dt, tb, fov

# ---Convert milliseconds in hour, minute, second--- #
def ms2time(ms):
    """
    Convert milliseconds in hour, minute, second

    .. warning::

       lost millisecond because second is integer

    .. warning::

       no day (but hour might be greater than 24)

    :param ms: millisecond

    :return: tuple (hour, minute, second)
    :rtype: tuple

    >>> # one hour
    >>> ms = 1000 * 60 * 60
    >>> (h, m, s) = ms2time(ms)
    >>> print ('{0} h {1} m {2} s'.format(h, m, s))
    1 h 0 m 0 s

    >>> # one hour + 1 ms
    >>> ms = (1000 * 60 * 60) + 1
    >>> (h, m, s) = ms2time(ms)
    >>> print ('{0} h {1} m {2} s'.format(h, m, s))
    1 h 0 m 0 s

    >>> # one day + one hour
    >>> ms = (1000 * 60 * 60 * 24) + (1000 * 60 * 60)
    >>> (h, m, s) = ms2time(ms)
    >>> print ('{0} h {1} m {2} s'.format(h, m, s))
    25 h 0 m 0 s
    """

    s = ms/1000
    m, s = divmod(s, 60)
    h, m = divmod(m, 60)
    # d, h = divmod(h, 24)
    return h, m, s

# ---Find coefficient of correction--- #
def find_corrcoef(dt, lon, lat, fov, ch):
    """Find the coefficient of correction
    *Inputs:
    dt  = Python date
    lat  = Latitude (degrees)
    lon  = Longitude (degrees)
    fov = Location in the swath
    ch  = list of extrated channel ('B1', 'B2', 'B3', 'B4', 'B5')

    *Outputs:
    coef = Coefficient of correction
    """

    # Find the day of year
    yday = array([d.timetuple().tm_yday for d in dt])

    coef = []
    # Read the coefficient correction file for each channel 
    for ich in ch:
        fic_corr = '/homedata/mleduc/CorrTb/corrTb_AMSUB_MHS_L1C_smooth_%s.nc' % ich
        fileID = netCDF4.Dataset(fic_corr)
        coef_corr = fileID.variables['corrTb'][:]
        if ich == ch[0]:
            lat_corr = fileID.variables['lat'][:]
            lon_corr = fileID.variables['lon'][:]
            day_corr = fileID.variables['day'][:]

            # Latitude limits of classification (ie tropical band)
            latstep = abs(diff(lat_corr))[0]
            latmin = lat_corr.min()-latstep/2.
            latmax = lat_corr.max()+latstep/2.

            # Geographic projection (lon/lat in degrees to x/y in m)
            projG = pyproj.Proj("+proj=merc +ellps=WGS84 +datum=WGS84")
            longd_corr, latgd_corr = meshgrid(lon_corr, lat_corr)
            xcorr, ycorr = projG(longd_corr, latgd_corr)

            # Kd-tree to find the nearest neighbour (euclidian distance)
            tree = spatial.cKDTree(zip(ravel(xcorr), ravel(ycorr)))

            # Find the day of correction
            nday = day_corr.size
            daystep = int(abs(diff(day_corr))[0])
            icorr = [[ii]*daystep for ii in xrange(nday)] + [0]*6

        fileID.close()
        # Only points in the area of classification
        inclass = logical_and(lat >= latmin, lat <= latmax)
        # index for days
        iday = hstack(icorr)[yday[inclass]-1]
        # index for location in the swath
        ifov = array(fov[inclass]-1, dtype=int)
        # index for lon/lat
        xx, yy = projG(lon[inclass], lat[inclass])
        _, ind = tree.query(zip(xx, yy))
        # Find coefficient of correction
        coefch = zeros(lat.shape, dtype=float32)+NaN
        coefch[inclass] = [coef_corr[dd, :, :, ff].flatten()[ii] for dd, ff, ii in zip(iday, ifov, ind)]
        # Concatenate channel
        coef.append(coefch)

    return float32(coef)

# ---New grid and time--- #
def newgrid(lonvec, latvec, lon, lat, dt, tb):
    """Find the coefficient of correction
    *Inputs:
    lonvec
    latvec
    lon
    lat
    dt
    tb

    *Outputs:
    tbm
    dtm
    """

    # from date to ordinal
    dtord = date2num(dt)

    # Find index of each pixel in new grid
    # Geographic projection (lon/lat in degrees to x/y in m)
    projG = pyproj.Proj("+proj=merc +ellps=WGS84 +datum=WGS84")
    longd, latgd = meshgrid(lonvec, latvec)
    xgd, ygd = projG(longd, latgd)
    xx, yy = projG(lon, lat)
    # Find the nearest neighbour (euclidian distance) with a kd-tree
    tree = spatial.cKDTree(zip(ravel(xgd), ravel(ygd)))
    _, indll = tree.query(zip(xx, yy))

    # Mean
    nch = tb.shape[0]
    tbm = zeros((nch, lonvec.size*latvec.size), dtype=float32)+NaN
    dtm = zeros((lonvec.size*latvec.size))+NaN
    for ii in unique(indll):
        tbm[:, ii] = nanmean(tb[:, (indll == ii)], axis=1)
        dtm[ii] = nanmean(dtord[(indll == ii)])

    return tbm, dtm
